;; "segs.scm" WB-tree File Based Associative String Data Base System.
;; Copyright (C) 1991, 1992, 1993, 2000, 2007 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place, Suite 330, Boston,
;; MA 02111, USA.
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of WB.
;;
;; The exception is that, if you link the WB library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the WB library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name WB.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; WB, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for WB, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;; routines in this file return success codes

(pragma.c "#include \"wbsys.h\"")

(java:import "Ents" "Blkio" "Prev" "Stats" "Blink" "Del" "Scan" "Wbdefs"
	     "Handle"
	     "Wbsys" "Blk" "BooleanMethods" "Pkt" "Lck" "Ent" "Seg" "Db" "Han")

;;@noindent
;;The @var{bsiz} of a segment (given in call to @code{make_seg})
;;is a parameter crucial for performance; balancing CPU time traversing
;;blocks with file-system latency.  @var{bsiz} should be an
;;integer multiple of the file-system block size.

;;@noindent
;;In the 1990s our nominal @var{bsiz} was 2.kiB; now it should
;;probably be 4.kiB, 8.kiB, or 16.kiB.

;; Segment not in use?
(define (seg-free? seg)
  (if (not lck-tab) (init-wb 75 150 2048))
  (cond ((or (negative? seg) (>= seg NUM-SEGS))
	 (edprintf "bad segment number %d\n" seg)
	 #f)
	((and (not (SEG:PORT seg))
	      (not (SEG:STR seg))
	      (zero? (SEG:USED seg)))
	 #t)
	(else #f)))

(define (check-seg? seg filename)
  (define file (blkio:open-read-only-file filename 128))
  (define tblk (make-bytes 25))
  (subbytes-read! tblk 0 25 file)
  (let ((allok? (and (zero? (str2long tblk 0))
		     (zero? (str2long tblk 4))
		     (zero? (str2long tblk 8))
		     (= #x77 (byte-ref tblk 23)) ;#\w
		     (= #x62 (byte-ref tblk 24))))) ;#\b
    (blkio:file-close file)
    allok?))

(define (errout reason-name seg filename)
  (edprintf "not a database %s %s\n" filename reason-name)
  (blkio:file-close (SEG:PORT seg))
  (do-seg-buffers
   seg
   #-(or SCM2JAVA SCM2CS) recycle-ent!
   #+SCM2JAVA (get-method "wb.Ents" "recycleEnt" (arg-class "Ent"))
   #+SCM2CS "recycleEnt"
   )
  (SEG:SET-PORT! seg #f)
  (SEG:SET-STR! seg #f)
  (SEG:SET-USED! seg 0)
  TYPERR)

;TBD - need to lck seg here.
;; Segment will be read-only if MODE is #f.

;;@body
;;@1 should be a non-negative number.  Opens the database @1 in @2 and
;;returns @1 if successful, a status code otherwise.  The database @1
;;will be read-only if the @3 argument is 0.  It will be read-write if
;;the @3 argument is 2.
(define (open-seg seg filename mode)
  (define bsiz (* 3 128))
  (if (zero? mode) (set! mode #f))
  (cond
   ((not (seg-free? seg))
    (edprintf "%s: segment in use %d\n" 'OPEN-SEG seg)
    ARGERR)
   ((> bsiz blk-size)
    ;;temporarily set bsiz so that we can get it from superblk
    (edprintf "unsupported %s %d > %d\n" 'bsiz bsiz blk-size)
    ARGERR)
   ((not (check-seg? seg filename)) -1)
   (else
    (let loop ((file (if mode (blkio:open-modify-file filename bsiz)
			 (blkio:open-read-only-file filename bsiz))))
      (cond
       ((if mode (output-port? file) (input-port? file))
	(SEG:SET-PORT! seg file)
	(SEG:SET-STR! seg filename)
	(SEG:SET-USED! seg 2)
	(SEG:SET-BSIZ! seg bsiz)
	(SEG:SET-FLC-LEN! seg (if mode -1 -2))
					;-1 means to read in "FLC" image.
					;-2 means read only.
	(let ((han (SEG:RT-HAN seg))
	      (tmp-str (make-bytes 5))) ;this should be longer
	  (cond
	   ((err? (bt-open seg 0 han (+ WCB-SAP WCB-SAR))) ; superblock
	    (errout "bt-open 0" seg filename))
	   ((not (eq? 2 (bt-get han BSIZ-byts 4 tmp-str)))
	    (errout "BSIZ" seg filename))
	   ((not (= bsiz (str2short tmp-str 0)))
	    (blkio:file-close file)
	    (set! bsiz (str2short tmp-str 0))
	    (cond
	     ((> bsiz blk-size) (errout "BSIZ too big." seg filename))
	     (else (loop (if mode (blkio:open-modify-file filename bsiz)
			     (blkio:open-read-only-file filename bsiz))))))
	   ((not (eq? 4 (bt-get han USED-byts 4 tmp-str)))
	    (errout "USED" seg filename))
	   (else
	    (SEG:SET-USED! seg (str2long tmp-str 0))
	    (cond ((not (eq? 5 (bt-get han FLD-byts 3 tmp-str)))
		   (errout "FLD" seg filename))
		  ((err? (bt-open seg (str2long tmp-str 1)
				  (SEG:FL-HAN seg) WCB-SAR))
		   (errout "FLC" seg filename))
		  (else
		   (if (not  (eqv? (HAN:TYP (SEG:FL-HAN seg)) FRL-TYP))
		       (dprintf "Older type freelist - still supported.\n"))
		   (HAN:SET-WCB! (SEG:FL-HAN seg) WCB-SAR)
		   seg))))))
       (else
	(if (if mode (input-port? file) (output-port? file))
	    (blkio:file-close file))
	(edprintf "could not open file %s\n" filename)
	IOERR))))))

;;@body
;;Closes database segment @1 and the file containing it.  If @2 is
;;NULL, then if there are any problems freeing buffers, then the close
;;is aborted.  A status code is returned.
(define (close-seg seg hammer?)
  (cond ((or (not (SEG:STR seg))
	     (zero? (SEG:USED seg)))
	 (edprintf "%s: segment %d already closed\n" 'CLOSE-SEG seg)
	 ARGERR)
	(else
	 (flush-flc! seg 5)
	 "leave only enough blocks to fit in flc in superblock."
	 (if (>= (SEG:FLC-LEN seg) 0)
	     (let* ((tmp-str (make-bytes 20)))
	       (do ((i (+ -1 (SEG:FLC-LEN seg)) (+ -1 i)))
		   ((negative? i))
		 (long2str! tmp-str (* 4 i) (vector-ref (SEG:FLC seg) i)))
	       (bt-put (SEG:RT-HAN seg) FLC-byts 3 tmp-str (* 4 (SEG:FLC-LEN seg)))
	       (SEG:SET-FLC-LEN! seg -1)))
	 (let ((ans
		(do-seg-buffers
		 seg
		 #-(or SCM2JAVA SCM2CS) flush-buffer
		 #+SCM2JAVA (get-method "wb.Ents" "flushBuffer" (arg-class "Ent")) 
		 #+SCM2CS	 "flushBuffer"
		 )))
	   (cond ((or (success? ans) hammer?)
		  (if (not (success? ans)) (set! ans NOTPRES))
		  (do-seg-buffers
		   seg
		   #-(or SCM2JAVA SCM2CS) purge-buffer
		   #+SCM2JAVA (get-method "wb.Ents" "purgeBuffer" (arg-class "Ent"))
		   #+SCM2CS "purgeBuffer"
		   )
		  (bt-close (SEG:RT-HAN seg))
		  (bt-close (SEG:FL-HAN seg))
		  (blkio:file-close (SEG:PORT seg))
		  (SEG:SET-PORT! seg #f)
		  (SEG:SET-STR! seg #f)
		  (SEG:SET-USED! seg 0)))
	   ans))))

(define db-version-str "wb2a1")
(define db-authors-str "A. Jaffer, J. Finger, R. Zito-Wolf")

;;@body
;;
;;The integer @1 must specify an unused segment.  The integer @3
;;specifies the size of B-tree blocks.  @3 should be an integer
;;multiple of the file-system block size.  Nominal value is 4096.
;;
;;@0 makes a new empty database named @2.  A status code is returned.
;;Call @cname{open-seg} to use the new database.
(define (make-seg seg filename bsiz)
  (cond
   ((or (not (seg-free? seg)) (not (try-lck? (SEG:LCK seg))))
    (edprintf "%s: segment in use %d\n" 'MAKE-SEG seg)
    ARGERR)
   ((> bsiz blk-size)
    (edprintf "unsupported %s %d > %d\n" 'bsiz bsiz blk-size)
    (unlck! (SEG:LCK seg))
    ARGERR)
   (else
    (let ((file (blkio:create-file filename bsiz)))
      (cond
       ((output-port? file)
	(SEG:SET-PORT! seg file)
	(SEG:SET-BSIZ! seg bsiz)
	(SEG:SET-USED! seg 3)
	(SEG:SET-STR! seg filename)
	(init-leaf-blk! empty-blk 0 DIR-TYP)
	(BLK:SET-TIME! empty-blk (current-time))
	(blkio:write file empty-blk bsiz 0)
	(init-leaf-blk! empty-blk 1 DIR-TYP)
	(BLK:SET-TIME! empty-blk (current-time))
	(blkio:write file empty-blk bsiz 1)
	(init-leaf-blk! empty-blk 2 FRL-TYP)
	(BLK:SET-TIME! empty-blk (current-time))
	(blkio:write file empty-blk bsiz 2)
	(blkio:file-close file)
	(set! file (blkio:open-modify-file filename bsiz))
	(cond ((output-port? file)
	       (SEG:SET-PORT! seg file)
	       (unlck! (SEG:LCK seg))
	       (let ((han (SEG:RT-HAN seg))
		     (tmp-str (make-bytes 5)))
		 (bt-open seg 0 han (+ WCB-SAP WCB-SAR))
		 (bt-put han NO-byts 0
			 db-version-str (string-length db-version-str))
		 (long2str! tmp-str 0 (SEG:USED seg))
		 (bt-put han USED-byts 4 tmp-str 4)
		 (short2str! tmp-str 0 (SEG:BSIZ seg))
		 (bt-put han BSIZ-byts 4 tmp-str 2)
		 (byte-set! tmp-str 0 4)
		 (long2str! tmp-str 1 1)
		 (bt-put han ROOT-byts 4 tmp-str 5)
		 (long2str! tmp-str 1 2)
		 (bt-put han FLD-byts 3 tmp-str 5)
		 (bt-put han FLC-byts 3 NO-byts 0)
		 (if (> bsiz 128)
		     (bt-put han AUTHORS-byts 7
			     db-authors-str (string-length db-authors-str)))
		 (close-seg seg (not #t)) ;don't close the segment if it is memory resident.
		 ))
	      (else (edprintf "couldn't open fresh file %s\n" filename)
		    (unlck! (SEG:LCK seg))
		    IOERR)))
       (else (edprintf "couldn't create new file %s\n" filename)
	     (unlck! (SEG:LCK seg))
	     IOERR))))))

;;@cindex wcb
;;@noindent The write-control-bits argument (@var{wcb}) to these
;;functions controls the latency of updates to the file after various
;;operations.  These bits are defined as follows:
;;
;;@multitable @columnfractions .1 .3 .6
;;@item @var{value}
;;@tab C-name
;;@tab Meaning
;;@item 1
;;@tab wcb_sap
;;@tab save block after PUTs
;;@cindex wcb_sap
;;@item 2
;;@tab wcb_sar
;;@tab save block after REMOVEs
;;@cindex wcb_sar
;;@item 4
;;@tab wcb_sac
;;@cindex wcb_sac
;;@tab force block save after cached block changes
;;@item 8
;;@tab wcb_fac
;;@tab flush buffer entirely after cached block changes (not currently
;;implemented)
;;@cindex wcb_fac
;;@end multitable

;;@body Opens bt-handle @3 to seg number @1, block number @2, and
;;returns the type of the block.  If no such block exists or is not a
;;root block, then a (negative) status code is returned.
(define (bt-open seg blk-num han wcb)
  (if (and (>= seg 0) (< seg NUM-SEGS) (SEG:STR seg)) ;allocated
      (let ((ent (get-ent seg blk-num ACCREAD)))
	(cond ((not ent) UNKERR)
	      ((not (root? (ENT:BLK ent)))
	       (release-ent! ent ACCREAD)
	       (edprintf "%s: not a root %d:%ld\n" 'BT-OPEN seg blk-num)
;;;	       (check-access!)
	       ARGERR)
	      (else
	       (HAN:SET-SEG! han seg)
	       (HAN:SET-NUM! han blk-num)
	       (HAN:SET-TYP! han (BLK:TYP (ENT:BLK ent))) ;TBD improve. (eh?)
	       (HAN:SET-LAST! han blk-num)
	       (if (BLK:TYP? (ENT:BLK ent) DIR-TYP)
		   (set! wcb (logior wcb (+ WCB-SAP WCB-SAR))))
	       (HAN:SET-WCB! han wcb)
	       (release-ent! ent ACCREAD)
;;;	       (check-access!)
	       (HAN:TYP han))))
      ARGERR))

;;@body Creates a new root block in seg @1 of type @2, opens bt-handle
;;@3 to it, and returns a status code.  If @1 has insufficient room to
;;create a new tree, then the @var{noroom} status code is returned.
;;
;;@0 can be used to create temporary b-trees.  Temporary trees will be
;;be reclaimed by check program after system crashes.  In order to
;;make a tree persistent, add it to a directory (tree).
(define (bt-create seg typ han wcb)
  (define ent (create-new-blk-ent seg))
  (cond ((not ent) NOROOM)
	(else (let* ((blk-num (ENT:ID ent)))
		(init-leaf-blk! (ENT:BLK ent) blk-num typ)
		(ENT:SET-DTY! ent #t)
		(ENT:SET-PUS! ent 0)
		(ents:ent-write ent)
		(HAN:SET-SEG! han seg)
		(HAN:SET-NUM! han blk-num)
		(HAN:SET-TYP! han typ)
		(HAN:SET-LAST! han blk-num)
		(if (eqv? typ DIR-TYP)
		    (set! wcb (logior wcb (+ WCB-SAP WCB-SAR))))
		(HAN:SET-WCB! han wcb)
		(release-ent! ent ACCWRITE)
;;;	     (check-access!)
		SUCCESS))))


;;@body Closes bt-handle @1 and returns @var{SUCCESS}.
;;
;;Currently, @0 has no effect other than to clear @1.
(define (bt-close han)
  (HAN:SET-SEG! han 0)
  (HAN:SET-NUM! han 0)
  (HAN:SET-TYP! han 0)
  (HAN:SET-LAST! han 0)
  SUCCESS)
